#if !defined(PARSER_HH)
#define PARSER_HH

#include <iostream>

#include "Lexer.hh"
#include "AstManager.hh"
#include "ErrorManager.hh"

///////////////////////////////////////////////////////////////////////////
class Parser {
///////////////////////////////////////////////////////////////////////////

public :

  Parser( Lexer& lex, AstManager& ast, ErrorManager& error ) : 
    lex_(lex), ast_(ast), error_(error) { }
  ~Parser( void ) { }
  
  // goal ::= expr
  AstNode* operator()( void ) {
    AstNode* expression = 0;
    // Advances the token cursor to the first token
    expression = parseExpression_();
    // Expects to find the token cursor on top of EOF
    if (lex_.cur().getPartOfSpeech() != Token::EOFF) {
      error_("Garbage after end of program");
    }
    return expression;
  }  
  
private :
  Lexer&        lex_;
  AstManager&   ast_;
  ErrorManager& error_;
  
  // expr ::= term expr'
  AstNode* parseExpression_() {
    AstNode* eterm = parseTerm_();
    AstNode* e = parseRestExpression_();
    if (e != 0) {
      e->setOperand(AstNode::LEFT, eterm);
    } else {
      e = eterm;
    }
    return e;
  }

  // expr' ::= + term expr' | - term expr' | \epsilon
  AstNode* parseRestExpression_() {
    AstNode* e = 0;
    if (isPlusOrMinus_()) {
      // ... if an operator is found, then there is a trailing expression, ...
      e = ast_.newOperator();
      e->setOpType(parseOpType_());
      // Parse term
      AstNode* eterm = parseTerm_();
      AstNode* erest = parseRestExpression_();
      if (erest != 0) {
	erest->setOperand(AstNode::LEFT, eterm);
	e->setOperand(AstNode::RIGHT, erest);
      } else {
	e->setOperand(AstNode::RIGHT, eterm);
      }
    }
    return e;
  }

  // term ::= factor term'
  AstNode* parseTerm_() {
    AstNode* efactor = parseFactor_();
    AstNode* e = parseRestTerm_();
    if (e != 0) {
      e->setOperand(AstNode::LEFT, efactor);
    } else {
      e = efactor;
    }
    return e;
  }

  // term' ::= * factor term' | / factor term' | \epsilon
  AstNode* parseRestTerm_() {
    AstNode* e = 0;
    if (isStarOrSlash_()) {
      // ... if an operator is found, then there is a trailing term, ...
      e = ast_.newOperator();
      e->setOpType(parseOpType_());
      // Parse term
      AstNode* efactor = parseFactor_();
      AstNode* erest = parseRestTerm_();
      if (erest != 0) {
	erest->setOperand(AstNode::LEFT, efactor);
	e->setOperand(AstNode::RIGHT, erest);
      } else {
	e->setOperand(AstNode::RIGHT, efactor);
      }
    }
    return e;
  }

  // factor ::= NUM | ( expr )
  AstNode* parseFactor_() {
    AstNode* e = 0;
    // Advance the token cursor
    lex_.next();
    if (lex_.cur().getPartOfSpeech() == Token::LPAR) {
      e = parseExpression_();
      if (lex_.cur().getPartOfSpeech() != Token::RPAR) {
	error_("Missing )");
      }
    } else if (lex_.cur().getPartOfSpeech() == Token::NUMBER) {
      e = ast_.newOperand();
      e->setValue(atoi(lex_.cur().getLexeme().data()));
    } else {
      error_("Factor expected");
    }
    // Advance the token cursor
    lex_.next();
    return e;
  }


  AstNode::OpType parseOpType_() {
    // Expects to find the token cursor on top of the operator
    char ch = lex_.cur().getLexeme()[0];
    switch (ch) {
      case '+' : return AstNode::PLUS;
      case '-' : return AstNode::MINUS;
      case '*' : return AstNode::TIMES;
      case '/' : return AstNode::FRAC;
    default :
      return AstNode::NOP;
    }
  }

  bool isPlusOrMinus_() {
    // Expects to find the token cursor on top of the operator
    char ch = lex_.cur().getLexeme()[0];
    switch (ch) {
    case '+' : 
    case '-' : return true;
    default :
      return false;
    }
  }

  bool isStarOrSlash_() {
    // Expects to find the token cursor on top of the operator
    char ch = lex_.cur().getLexeme()[0];
    switch (ch) {
    case '*' : 
    case '/' : return true;
    default :
      return false;
    }
  }

}; // End of class Parser

#endif
